// (c) 2022 cepstrum.co.jp           //
// This is not free software.        //
//
// conventional FIR filter & LMS algorithm demo
//
// suppress random noise by adaptive filter
//
//  x ---+------------------------------>(+)---+---> e
//       |                            +   ^    |
//       |                         ^      | -  |
//       |                         |      |    |
//       |    +-------+ zx      +-----+ y |    |
//       +--->| delay +----+--->| FIR +---+----|---> output
//            +-------+    |    +-----+        |
//            z^(-100)     |       |           |
//                         |    +-----+        |
//                         +--->| LMS |<-------+
//                              +-----+
//

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <math.h>
#include <time.h>

#define MU      5e-7
#define ITERATE 300000
#define FIRLEN  100000
#define PERIOD  1499

float  w[FIRLEN];
float  xbuf[FIRLEN];
float  out1[ITERATE];
float  out2[ITERATE];
float  out3[ITERATE];
float  y;
float  x;
float  zx;
float  e;

time_t start_time;
time_t finish_time;

int    i;
char   c;

//--------------------------------------------
// conventional FIR filter

float fir(float x, float xbuf[], float coef[], int length) {
  int   ix;
  float y;

  for (ix=length-1; 1<=ix; ix=ix-1) xbuf[ix]=xbuf[ix-1];
  xbuf[0]=x;
  y=0.0;
  for (ix=0; ix<length; ix=ix+1) y=y+xbuf[ix]*coef[ix];
  return y;
}

//--------------------------------------------
// conventional LMS algorithm

void lms(float w[], float e, float xbuf[], int length) {
  int ix;

  for (ix=0; ix<length; ix=ix+1) w[ix]=w[ix]+e*xbuf[ix];
}

//--------------------------------------------
// 100 sample delay (ring buffer)

float delay100z(float x) {
  static unsigned ix=0;
  static float    buf[100]={
                             0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                             0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                             0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                             0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                             0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                             0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                             0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                             0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                             0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                             0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0
                           };
  float           y;

  y=buf[ix];
  buf[ix]=x;
  ix=(ix+1)%100;
  return y;
}

//--------------------------------------------

int main() {

  for (i=0; i<FIRLEN; i=i+1) {
    w[i]=0.0;
    xbuf[i]=0.0;
  }
  for (i=0; i<ITERATE; i=i+1) {
    out1[i]=0.0;
    out2[i]=0.0;
    out3[i]=0.0;
  }

  fprintf(stderr, "running %iloop %itap adaptive filter (FIR filter + LMS algorithm)\n", ITERATE, FIRLEN);
  fprintf(stderr, "please wait...\n");

  start_time=time(NULL);
  zx=0.0;
  for (i=0; i<ITERATE; i=i+1) {
    x=(i%PERIOD)/(float)PERIOD-0.5;             // generate periodic signal (saw tooth)
    x=x+(float)rand()/(float)RAND_MAX-0.5;      // add random noise
    zx=delay100z(x);
    y=fir(zx, xbuf, w, FIRLEN);
    e=x-y;
    lms(w, 2.0*MU*e, xbuf, FIRLEN);
    out1[i]=x;
    out2[i]=y;
    out3[i]=w[1401];
    if (i%(int)(ITERATE/75)==0) {
      fprintf(stderr, "*");
    }
  }
  fprintf(stderr, "\n");
  finish_time=time(NULL);

  fprintf(stderr, "%i[sec]\n", (int)difftime(finish_time, start_time));
  fprintf(stderr, "hit retun key to proceed or ctrl-C to exit");
  scanf("%c", &c);

  //for (i=0; i<FIRLEN; i=i+1) printf("%7i %15.12f\n", i, w[i]);
  for (i=0; i<ITERATE; i=i+1) printf("%15.12f\t%15.12f\n", out1[i], out2[i]);
  //for (i=0; i<ITERATE; i=i+1) printf("%15.12f\n", out3[i]);

}


